/*
 * mysql_auth_md5crypt - additional encryption for mysql_auth
 *
 * mysql_auth_md5cpyt.c
 * (C) 2008 Andras Miko
 * Released under GPL, see COPYING-2.0 for details.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h> /* required for random */
#include "define.h"
#include "md5.h"
#include "mysql_auth_md5.h"

int substr (char *buffer, const char *pstr, int start, int numchars) {
//	char *pnew = malloc ((numchars + 1) * sizeof (char));
	char pnew[numchars];
	strncpy (pnew, pstr + start, numchars);
	pnew[numchars] = '\0';
	strcpy (buffer, pnew);
	return 0;
}

int hex2bin (char *buffer, const char *str) {                                             // function hex2bin ($str);
	char *tmp = malloc (strlen (str) * sizeof (char));
	char *buf2;
	char *nstr = malloc ((strlen (str) * 2 + 1) * sizeof (char));
	char *ret = malloc ((strlen (str) * 2 + 1) * sizeof (char));
	if (tmp == NULL || nstr == NULL || ret == NULL) exit(1);
	int h, i, j = 0;
	int len = strlen (str);                                                                 // $len = strlen ($str);
	tmp[0] = '\0';
	nstr[0] = '\0';
	ret[0] = '\0';
	for (i = 0; i < len; i += 2) {                                                          // for ($i = 0; $i < $len; $i += 2) {
		substr (tmp, str, i, 2);                                                            // <buffer> = substr ($str, $i, 2)
		sscanf (tmp, "%x", &h);                                                             // sscanf (<buffer>, "%x");
		ret[j] = h;                                                                         // $nstr .= chr ($num[0]);
		j++;
	}                                                                                       // }
	memcpy (buffer, ret, j);
	buffer[j] = '\0';
	free (tmp);
	free (nstr);
	free (ret);
	return 0;                                                                               // return $nstr;
}                                                                                           // }

int to64 (char *buffer, int v, int n) {                                                     // function to64 ($v, $n) {
	char *ret = malloc (6 * sizeof (char));                                                 // $ret = "";
	char ITOA64[65];
	int i = 0;
	int v2;
	strcpy (ITOA64, "./0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz");    // $ITOA64 = "./0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
	while ((n - 1) >= 0) {                                                                  // while (($n - 1) >= 0) {
		n--;                                                                                // $n--;
		v2 = v & 0x3f;                                                                      // <buffer>= $v & 0x3f;
		ret[i] = ITOA64[v2];                                                                // $ret .= $ITOA64[<buffer>];
		i++;
		v = v >> 6;                                                                         // $v = $v >> 6;
	}                                                                                       // }
	memcpy (buffer, ret, i);
	buffer[i] = '\0';
	free(ret);
	return 0;                                                                               // return $ret;
}

int create_salt (char *buffer) {                                                            // function create_salt () {
	char *salt = malloc (33 * sizeof (char));
	char *cRnd = malloc (15 * sizeof (char));
	salt[0] = '\0';
	cRnd[0] = '\0';
	srand (time(0));                                                                        // srand ((double) microtime ()*1000000);
	sprintf (cRnd, "%d", rand ());                                                          // <buffer1> = rand (0,9999999)
	f_md5 (salt, cRnd);
	substr (salt, salt, 0, 8);                                                      // $salt = substr (md5 (<buffer1>), 0, 8);
	strcpy (buffer, salt);
	free(salt);
	free(cRnd);
	return 0;                                                                               // return $salt;
}                                                                                           // }

void md5crypt (char *retval, const char *cpw, const char *csalt, const char *cmagic) {      // function md5crypt ($pw, $salt ="", $magic ="") {
	char *ret = malloc (100 * sizeof (char));
	char *MAGIC = "$1$";                                                                    // $MAGIC = "$1$";
	char *pw = malloc (150 * sizeof (char));
	char *salt = malloc (16 * sizeof (char));
	char *magic = malloc (50 * sizeof (char));
	char *final = malloc (200 * sizeof (char));
	char *ctx = malloc (200 * sizeof (char));
	char *buffer = malloc (100 * sizeof (char));
	char *h2b = malloc (350 * sizeof (char));
	char *tt64 = malloc (6 * sizeof (char));
	if (ret == NULL || pw == NULL || salt == NULL || magic == NULL || final == NULL || ctx == NULL || buffer == NULL || h2b == NULL) exit(1);
	int final_ptr = 0;
	int i = 0;
	ctx[0] = '\0';
	int ctx_ptr = 0;

	strcpy (pw, cpw);
	if (strlen (csalt) == 0) create_salt (salt);                                            // if ($salt == "") $salt = create_salt ();
		else strcpy (salt, csalt);
	if (strlen (cmagic) == 0) strcpy (magic, MAGIC);                                        // if ($magic == "") $magic = $MAGIC;
		else strcpy (magic, cmagic);
	substr (salt, salt, 0, 8);                                                              // $salt = substr ($salt, 0, 8);
	strcpy (ctx, pw);                                                                       // $ctx = $pw . $magic . $salt;
	strcat (ctx, magic);                                                                    // $ctx = $pw . $magic . $salt;
	strcat (ctx, salt);                                                                     // $ctx = $pw . $magic . $salt;
	ctx_ptr = strlen (pw) + strlen (magic) + strlen (salt);
	strcpy (final, pw);
	strcat (final, salt);
	strcat (final, pw);
	final_ptr = strlen (pw) + strlen (salt) + strlen (pw);
	f_md5 (h2b, final);
	hex2bin (h2b, h2b);
	memcpy (final, h2b, final_ptr);

	for (i = strlen (pw); i > 0; i -= 16) {
		if (i > 16) {
			memcpy (ctx + ctx_ptr, final, 16);
			ctx_ptr += 16;
		} else {
			memcpy (ctx + ctx_ptr, final, i);
			ctx_ptr += i;
		}
	}
	ctx[ctx_ptr] = '\0';
	i = strlen (pw);
	int j = 0;
	while (i > 0) {
		if (i & 1) {
			ctx[ctx_ptr] = (char)0;
		}
		else {
			ctx[ctx_ptr] = pw[0];
		}
		ctx_ptr++;
		i = i >> 1;
	}
	ctx[ctx_ptr] = '\0';
	// begin for md5
	md5_state_t state;
	md5_byte_t digest[16];
	int di;
	char hex_output[16 * 2 + 1];
	// end for md5
	md5_init (&state);
	md5_append (&state, (const md5_byte_t *)ctx, ctx_ptr);
	md5_finish (&state, digest);
	for (di = 0; di < 16; ++di) sprintf (hex_output + di * 2, "%02x", digest[di]);

	hex2bin (h2b, hex_output);
	memcpy (final, h2b, strlen (hex_output) / 2);
	memcpy (final + strlen (hex_output) / 2, MAGIC + strlen (MAGIC), 1);

	char * ctx1 = malloc (100 * sizeof (char));
	if (ctx1 == NULL) exit(1);
	for (i = 0; i < 1000; i++) {
		int ptr = 0;
		ctx1[0] = '\0';
		if (i & 1) {
			memcpy (ctx1 + ptr, pw, strlen (pw));
			ptr += strlen (pw);
		} else {
			memcpy (ctx1 + ptr, final, 16);
			ptr += 16;
		}
		if (i % 3) {
			memcpy (ctx1 + ptr, salt, 8);
			ptr += 8;
		}
		if (i % 7) {
			memcpy (ctx1 + ptr, pw, strlen (pw));
			ptr += strlen (pw);
		}
		if (i & 1) {
			memcpy (ctx1 + ptr, final, 16);
			ptr += 16;
		} else {
			memcpy (ctx1 + ptr, pw, strlen (pw));
			ptr += strlen (pw);
		}
		md5_init (&state);
		md5_append (&state, (const md5_byte_t *)ctx1, ptr);
		md5_finish (&state, digest);
		for (di = 0; di < 16; ++di) sprintf (hex_output + di * 2, "%02x", digest[di]);

		hex2bin (h2b, hex_output);
		memcpy (final, h2b, strlen (hex_output) / 2);
	}
	char * passwd = malloc (30 * sizeof (char));
	if (passwd == NULL) exit (1);
	passwd[0] = '\0';
	unsigned long int ord, ord1, ord2, ord3;

	ord1 = (char) final[0];
	ord1 = ord1 & 0xff;
	ord1 = ord1 << 16;
	ord2 = (char) final[6];
	ord2 = ord2 & 0xff;
	ord2 = ord2 << 8;
	ord3 = (char) final[12];
	ord3 = ord3 & 0xff;
	ord = ord1 | ord2 | ord3;
	to64 (tt64, ord, 4);
	strcpy (passwd, tt64);

	ord1 = (char) final[1];
	ord1 = ord1 & 0xff;
	ord1 = ord1 << 16;
	ord2 = (char) final[7];
	ord2 = ord2 & 0xff;
	ord2 = ord2 << 8;
	ord3 = (char) final[13];
	ord3 = ord3 & 0xff;
	ord = ord1 | ord2 | ord3;
	to64 (tt64, ord, 4);
	strcat (passwd, tt64);

	ord1 = (char) final[2];
	ord1 = ord1 & 0xff;
	ord1 = ord1 << 16;
	ord2 = (char) final[8];
	ord2 = ord2 & 0xff;
	ord2 = ord2 << 8;
	ord3 = (char) final[14];
	ord3 = ord3 & 0xff;
	ord = ord1 | ord2 | ord3;
	to64 (tt64, ord, 4);
	strcat (passwd, tt64);

	ord1 = (char) final[3];
	ord1 = ord1 & 0xff;
	ord1 = ord1 << 16;
	ord2 = (char) final[9];
	ord2 = ord2 & 0xff;
	ord2 = ord2 << 8;
	ord3 = (char) final[15];
	ord3 = ord3 & 0xff;
	ord = ord1 | ord2 | ord3;
	to64 (tt64, ord, 4);
	strcat (passwd, tt64);

	ord1 = (char) final[4];
	ord1 = ord1 & 0xff;
	ord1 = ord1 << 16;
	ord2 = (char) final[10];
	ord2 = ord2 & 0xff;
	ord2 = ord2 << 8;
	ord3 = (char) final[5];
	ord3 = ord3 & 0xff;
	ord = ord1 | ord2 | ord3;
	to64 (tt64, ord, 4);
	strcat (passwd, tt64);

	ord1 = (char) final[11];
	ord1 = ord1 & 0xff;
	to64 (tt64, ord1, 2);
	strcat (passwd, tt64);

	magic = strcat (magic, salt);
	magic = strcat (magic, "$");
	magic = strcat (magic, passwd);

	strcpy (retval, magic);
	free(ret);
	free(pw);
	free(salt);
	free(magic);
	free(final);
	free(ctx);
	free(buffer);
	free(h2b);
	free(tt64);
	free(ctx1);
	free(passwd);
	return;
}                                                                                           // }
